#include <windows.h>

#include <io.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vector>

#include <GL/glew.h>
#include <GL/wglew.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <math.h>

const int screen_width = 64;
const int screen_height = 64;

struct _Screen
{
	int width;
	int height;
	unsigned char* buffer;
	unsigned int texture;
	_Screen() : width(screen_width), height(screen_height), buffer(NULL), texture(0) {}
};

bool CreateScreen(const int width, const int height, _Screen* screen);
void DeleteScreen(_Screen* screen);
bool LoadVxl(char* filename);
void DrawVxlRaycast();


void FormPaint(HWND Handle, HGLRC hrc);
void FormResize(HWND Handle, int ClientWidth, int ClientHeight, HGLRC hrc);
void Init();
void FormKeyDown(HWND Handle, HGLRC hglrc, WPARAM Key);

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
BOOL setupPixelFormat(HDC hdc);

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, INT iCmdShow)
{
	HWND		hwnd;
	HACCEL		hAccel;
	MSG		msg;
	WNDCLASS	wndclass;
	char		szAppName[] = TEXT("Voxel");

	wndclass.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wndclass.lpfnWndProc = WndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = hInstance;
	wndclass.hIcon = LoadIcon(hInstance, "engine");
	wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground = NULL;
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = szAppName;

	if (!RegisterClass(&wndclass))
	{
		MessageBox(NULL, TEXT("Unable to register window class."), szAppName, MB_ICONERROR);
		return 0;
	}
	hwnd = CreateWindow(szAppName,			// window class name
		TEXT("Voxel"),		// window caption
		WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,		// window style
		CW_USEDEFAULT,			// initial x position
		CW_USEDEFAULT,			// initial y position
		CW_USEDEFAULT,			// initial x size
		CW_USEDEFAULT,			// initial y size
		NULL,				// parent window handle
		NULL,				// window menu handle
		hInstance,			// program instance handle
		NULL);				// creation parameters

	hAccel = LoadAccelerators(hInstance, szAppName);
	ShowWindow(hwnd, iCmdShow);
	UpdateWindow(hwnd);

	while (TRUE)
	{
		if (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
		{
			if (GetMessage(&msg, NULL, 0, 0) > 0)
			{
				if (!TranslateAccelerator(hwnd, hAccel, &msg))
				{
					TranslateMessage(&msg);
					DispatchMessage(&msg);
				}
			}
			else {
				break;
			}
		}
		else
		{
//			EngineProc(E_RENDER, 0, 0);
		}
	}

	if (!UnregisterClass(szAppName, hInstance))
	{
		MessageBox(NULL, TEXT("Unable to unregister class."), szAppName, MB_ICONERROR);
		return 1;
	}

	return msg.wParam;
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static HDC		hdc;
	static HGLRC		hglrc;
	static int cxClient;
	static int cyClient;
	static RECT rect;

	switch (message)
	{
	case WM_CREATE:
		hdc = GetDC(hwnd);
		setupPixelFormat(hdc);
		hglrc = wglCreateContext(hdc);
		wglMakeCurrent(hdc, hglrc);
		glewInit();



		Init();

		if (false == LoadVxl("./data/seed7t1.vxl"))
		{
			MessageBox(hwnd, "Error loading VXL.\n", "Error", 0);
			return 0;
		}

		return 0;

	case WM_MOUSEMOVE:
	{
		int	x, y;

		x = LOWORD(lParam);
		y = HIWORD(lParam);
	}
	return 0;

	case WM_TIMER:
	{
		int width = cxClient;
		int height = cyClient;

		InvalidateRect(hwnd, &rect, FALSE);
		return 0;
	}

	case WM_KEYDOWN:
		FormKeyDown(hwnd, hglrc, wParam);
		return 0;

	case WM_SIZE:
	{
		int	width, height;

		width = LOWORD(lParam);
		height = HIWORD(lParam);

		cxClient = width;
		cyClient = height;
		rect.right = cxClient;
		rect.bottom = cyClient;


		FormResize(hwnd, width, height, hglrc);
	}
	return 0;
	case WM_PAINT:
		FormPaint(hwnd, hglrc);
		return 0;

	case WM_SYSCOMMAND:
		switch (wParam)
		{
		case SC_SCREENSAVE:
		case SC_MONITORPOWER:
			return 0;
		}
		break;

	case WM_DESTROY:
		if (hglrc)
		{
			wglMakeCurrent(NULL, NULL);
			wglDeleteContext(hglrc);
			ReleaseDC(hwnd, hdc);
		}
		PostQuitMessage(0);
		return 0;
	}
	return DefWindowProc(hwnd, message, wParam, lParam);
}

BOOL setupPixelFormat(HDC hdc)
{
	PIXELFORMATDESCRIPTOR pfd;
	INT pixelformat;

	pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_GENERIC_ACCELERATED | PFD_DOUBLEBUFFER;
	pfd.dwLayerMask = PFD_MAIN_PLANE;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 32;
	pfd.cDepthBits = 24;
	pfd.cAccumBits = 32;
	pfd.cStencilBits = 32;

	if ((pixelformat = ChoosePixelFormat(hdc, &pfd)) == 0)
	{
		MessageBox(NULL, "ChoosePixelFormat failed", "Error", MB_OK);
		return FALSE;
	}

	if (SetPixelFormat(hdc, pixelformat, &pfd) == FALSE)
	{
		MessageBox(NULL, "SetPixelFormat failed", "Error", MB_OK);
		return FALSE;
	}
	return TRUE;
}


#ifndef M_PI
#define M_PI 3.1415926535
#endif

typedef struct 
{
	GLfloat x, y, z, w;
} Vector3D;

typedef struct 
{
	GLfloat m[16];
} Matrix;

typedef struct 
{
	Vector3D eye;
	Vector3D center;
	Vector3D up;
	Vector3D side;
} Camera;

Matrix Identity = {	1.0, 0.0, 0.0, 0.0,
					0.0, 1.0, 0.0, 0.0,
					0.0, 0.0, 1.0, 0.0,
					0.0, 0.0, 0.0, 1.0};

Camera default_cam = {	{0.0, 0.0, 2.0, 1.0},
						{0.0, 0.0, 1.0, 1.0},
						{0.0, 1.0, 2.0, 1.0},
						{1.0, 0.0, 2.0, 1.0}};

Camera *player = &default_cam;

_Screen screen;

bool fire = false;

GLfloat stepx = 0.5,
	stepy = 0.5,
	stepz = 0.5,
	degleft = 10.0,
	degup	= 10.0,
	degroll = 10.0;

const GLfloat h = 0.05f; // voxel grid cell size
const unsigned long kFar = 2;
const unsigned long kRange = 10;
const GLfloat d1 = 4.0;
const GLfloat d2 = 4.0;
const GLfloat ddx = 1.0 / screen_width;
const GLfloat ddy = 1.0 / screen_height;
const Vector3D boardorigin  = {-512 * h, -512 * h, -128 * h, 1.0};


inline float DotProduct(const Vector3D &v1, const Vector3D &v2)
{
	return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}

Vector3D CrossProduct(const Vector3D &v1, const Vector3D &v2)
{
	Vector3D result;
	result.x = v1.y*v2.z - v1.z*v2.y;
	result.y = v1.z*v2.x - v1.x*v2.z;
	result.z = v1.x*v2.y - v1.y*v2.x;
	return result;
}

inline float Length(const Vector3D &v)
{
	return sqrt(v.x*v.x + v.y*v.y + v.z*v.z);
}

Vector3D Normalize(const Vector3D &v)
{
	Vector3D result = v;
	float length = Length(v);
	if(length)
	{
		result.x /= length;
		result.y /= length;
		result.z /= length;
	}
	return result;
}

Vector3D TransformVector(const Vector3D &v, const Matrix &m)
{
	Vector3D result;
	result.x = (m.m[0] * v.x) + (m.m[4] * v.y) + (m.m[8] * v.z) + (m.m[12] * v.w);
	result.y = (m.m[1] * v.x) + (m.m[5] * v.y) + (m.m[9] * v.z) + (m.m[13] * v.w);
	result.z = (m.m[2] * v.x) + (m.m[6] * v.y) + (m.m[10] * v.z) + (m.m[14] * v.w);
	result.w = (m.m[3] * v.x) + (m.m[7] * v.y) + (m.m[11] * v.z) + (m.m[15] * v.w);
	return result;
}

//Rodrigues' rotation formula
Matrix Rotate(float degrees, const Vector3D& axis)
{
	Matrix I, Q, S, R;
	unsigned int i;
	float radians = (degrees * M_PI) / 180.0;

	R = S = Q = I = Identity;

	Q.m[0] = axis.x*axis.x; Q.m[4] = axis.x*axis.y; Q.m[8] = axis.x*axis.z;
	Q.m[1] = axis.x*axis.y; Q.m[5] = axis.y*axis.y; Q.m[9] = axis.y*axis.z;
	Q.m[2] = axis.x*axis.z; Q.m[6] = axis.y*axis.z; Q.m[10] = axis.z*axis.z;

	S.m[0] =  0.0;
	S.m[4] = -axis.z;
	S.m[8] =  axis.y;
	S.m[1] =  axis.z;
	S.m[5] =  0.0;
	S.m[9] = -axis.x;
	S.m[2] = -axis.y;
	S.m[6] =  axis.x;
	S.m[10] = 0.0;

	for(i=0; i<12; ++i)
	{
		if(i==3 || i==7 || i==11)
			continue;
		I.m[i] *= cos(radians);
		Q.m[i] *= (1 - cos(radians));
		S.m[i] *= sin(radians);
		R.m[i] = I.m[i] + Q.m[i] + S.m[i];
	}

	return R;
}

Matrix Translate(float x, float y, float z)
{
	Matrix T = Identity;

	T.m[12] = x;
	T.m[13] = y;
	T.m[14] = z;
	T.m[15] = 1.0;

	return T;
}

void Init()
{
	glClearColor(0.0, 0.0, 0.0, 1.0);
	glEnable(GL_DEPTH_TEST);
	CreateScreen(screen_width, screen_height, &screen);
}


void DrawAxes()
{
	glLineWidth(5);

	glColor3f(0.0, 1.0, 0.0);  //X - Green
	glBegin(GL_LINES);
		glVertex3d( 1.0, 0.0, 0.0);
		glVertex3d(-1.0, 0.0, 0.0);

		glVertex3d( 1.0, 0.0, 0.0);
		glVertex3d(0.75,  0.025, 0.0);
		glVertex3d( 1.0, 0.0, 0.0);
		glVertex3d(0.75, -0.025, 0.0);

		glVertex3d( 1.0, 0.0, 0.0);
		glVertex3d(0.75, 0.0,  0.025);
		glVertex3d( 1.0, 0.0, 0.0);
		glVertex3d(0.75, 0.0, -0.025);
	glEnd();

	glColor3f(1, 1, 0);  //Y - Yellow
	glBegin(GL_LINES);
		glVertex3d(0.0,  1.0, 0.0);
		glVertex3d(0.0, -1.0, 0.0);

		glVertex3d(0.0,  1.0, 0.0);
		glVertex3d(0.0, 0.75,  0.025);
		glVertex3d(0.0,  1.0, 0.0);
		glVertex3d(0.0, 0.75, -0.025);

		glVertex3d(0.0,  1.0, 0.0);
		glVertex3d( 0.025, 0.75, 0.0);
		glVertex3d( 0.0,  1.0, 0.0);
		glVertex3d(-0.025, 0.75, 0.0);
	glEnd();

	glColor3f(1, 0, 0);  //Z - Red
	glBegin(GL_LINES);
		glVertex3d(0.0, 0.0,  1.0);
		glVertex3d(0.0, 0.0, -1.0);

		glVertex3d( 0.0, 0.0,  1.0);
		glVertex3d( 0.025, 0.0, 0.75);
		glVertex3d( 0.0, 0.0,  1.0);
		glVertex3d(-0.025, 0.0, 0.75);

		glVertex3d(0.0,  0.0,  1.0);
		glVertex3d(0.0,  0.025, 0.75);
		glVertex3d(0.0,  0.0,  1.0);
		glVertex3d(0.0, -0.025, 0.75);
	glEnd();
}

void DrawCentroids()
{
	Vector3D up;
	up.x = player->up.x - player->eye.x;
	up.y = player->up.y - player->eye.y;
	up.z = player->up.z - player->eye.z;

	Vector3D front;
	front.x = player->center.x - player->eye.x;
	front.y = player->center.y - player->eye.y;
	front.z = player->center.z - player->eye.z;

	Vector3D right;
	right.x = player->side.x - player->eye.x;
	right.y = player->side.y - player->eye.y;
	right.z = player->side.z - player->eye.z;

	const GLfloat d1 = 0.02;
	Vector3D origin;
	origin.x = player->eye.x + d1 * front.x;
	origin.y = player->eye.y + d1 * front.y;
	origin.z = player->eye.z + d1 * front.z;
	origin.w = 1.0;

	const GLfloat d2 = 0.5;
	glLineWidth(5.0);
	glBegin(GL_LINES);
		glColor3f(1.0, 1.0, 1.0);
		glVertex3f(origin.x, origin.y, origin.z);
		glColor3f(0.4, 0.4, 0.6);
		glVertex3f( origin.x + d2 * front.x
			, origin.y + d2 * front.y
			, origin.z + d2 * front.z
		);
	glEnd();
	glBegin(GL_LINES);
		glColor3f(1.0, 1.0, 1.0);
		glVertex3f(origin.x, origin.y, origin.z);
		glColor3f(0.4, 0.6, 0.4);
		glVertex3f( origin.x + d2 * right.x
			, origin.y + d2 * right.y
			, origin.z + d2 * right.z
			);
	glEnd();
	glBegin(GL_LINES);
		glColor3f(1.0, 1.0, 1.0);
		glVertex3f(origin.x, origin.y, origin.z);
		glColor3f(0.6, 0.4, 0.4);
		glVertex3f( origin.x + d2 * up.x
			, origin.y + d2 * up.y
			, origin.z + d2 * up.z
			);
	glEnd();
}

void DrawLaserFire()
{
	glColor3f(1.0, 0.0, 0.0);
	glLineWidth(5.0f);
	glBegin(GL_LINES);
		glVertex3f(-1.0, 1.0, -1.0);
		glVertex3f(-0.025, 0.025, -1.0);
		glVertex3f( 1.0, 1.0, -1.0);
		glVertex3f( 0.025, 0.025, -1.0);
		glVertex3f( 1.0, -1.0, -1.0);
		glVertex3f( 0.025, -0.025, -1.0);
		glVertex3f(-1.0, -1.0, -1.0);
		glVertex3f(-0.025, -0.025, -1.0);
	glEnd();
}

void Draw()
{
	glClear(GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	DrawVxlRaycast();
	if (true == fire)
		DrawLaserFire();
	glFlush();
}

void FormPaint(HWND Handle, HGLRC hrc)
{
	PAINTSTRUCT ps;

	HDC DC = BeginPaint(Handle, &ps);
	Draw();
	SwapBuffers(DC);
	EndPaint(Handle, &ps);
}

void FormResize(HWND Handle, int ClientWidth, int ClientHeight, HGLRC hrc)
{
	glViewport(0, 0, (GLsizei)ClientWidth, (GLsizei)ClientHeight);
	glPushMatrix();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glPopMatrix();
	InvalidateRect(Handle, NULL, false);
	FormPaint(Handle, hrc);
}


// ============================================================
// FPS-style 6DOF camera originated from Crystal Ball interface
// ============================================================
void MoveZ(float step, Vector3D *eye, Vector3D *center, Vector3D *up, Vector3D *side)
{
	Vector3D frontaxis;
	frontaxis.x = center->x - eye->x;
	frontaxis.y = center->y - eye->y;
	frontaxis.z = center->z - eye->z;
	frontaxis = Normalize(frontaxis);

	eye->x += step * frontaxis.x;
	eye->y += step * frontaxis.y;
	eye->z += step * frontaxis.z;
	center->x += step * frontaxis.x;
	center->y += step * frontaxis.y;
	center->z += step * frontaxis.z;
	up->x += step * frontaxis.x;
	up->y += step * frontaxis.y;
	up->z += step * frontaxis.z;
	side->x += step * frontaxis.x;
	side->y += step * frontaxis.y;
	side->z += step * frontaxis.z;
}

void MoveY(float step, Vector3D *eye, Vector3D *center, Vector3D *up, Vector3D *side)
{
	Vector3D upaxis;
	upaxis.x = up->x - eye->x;
	upaxis.y = up->y - eye->y;
	upaxis.z = up->z - eye->z;
	upaxis = Normalize(upaxis);

	eye->x += step * upaxis.x;
	eye->y += step * upaxis.y;
	eye->z += step * upaxis.z;
	center->x += step * upaxis.x;
	center->y += step * upaxis.y;
	center->z += step * upaxis.z;
	up->x += step * upaxis.x;
	up->y += step * upaxis.y;
	up->z += step * upaxis.z;
	side->x += step * upaxis.x;
	side->y += step * upaxis.y;
	side->z += step * upaxis.z;
}

void MoveX(float step, Vector3D *eye, Vector3D *center, Vector3D *up, Vector3D *side)
{
	Vector3D sideaxis;
	sideaxis.x = side->x - eye->x;
	sideaxis.y = side->y - eye->y;
	sideaxis.z = side->z - eye->z;
	sideaxis = Normalize(sideaxis);

	eye->x += step * sideaxis.x;
	eye->y += step * sideaxis.y;
	eye->z += step * sideaxis.z;
	center->x += step * sideaxis.x;
	center->y += step * sideaxis.y;
	center->z += step * sideaxis.z;
	up->x += step * sideaxis.x;
	up->y += step * sideaxis.y;
	up->z += step * sideaxis.z;
	side->x += step * sideaxis.x;
	side->y += step * sideaxis.y;
	side->z += step * sideaxis.z;
}

void TurnLeft(float degrees, Vector3D *eye, Vector3D *center, Vector3D *up, Vector3D *side)
{
	//In Crystal Ball interface:
	//eye = eye * rotate(degrees, normalize(up));
	//*eye = TransformVector(*eye, Rotate(degrees, Normalize(up)));

	//Set up axis to rotate around. It's origin will be the origin of world coordinates.
	Vector3D upaxis;
	upaxis.x = up->x - eye->x; //0.0
	upaxis.y = up->y - eye->y; //1.0
	upaxis.z = up->z - eye->z; //0.0
	upaxis = Normalize(upaxis);

	Vector3D oldeye;
	oldeye.x = eye->x;
	oldeye.y = eye->y;
	oldeye.z = eye->z;

	//Translate to the origin of world coordinates
	center->x -= eye->x;
	center->y -= eye->y;
	center->z -= eye->z;
	up->x -= eye->x;
	up->y -= eye->y;
	up->z -= eye->z;
	side->x -= eye->x;
	side->y -= eye->y;
	side->z -= eye->z;
	eye->x -= eye->x; //
	eye->y -= eye->y; // :)
	eye->z -= eye->z; //

	//Transform: rotate
	*eye = TransformVector(*eye, Rotate(degrees, upaxis));
	*center = TransformVector(*center, Rotate(degrees, upaxis));
	*up = TransformVector( *up, Rotate(degrees, upaxis));
	*side = TransformVector(  *side, Rotate(degrees, upaxis));

	//Translate back to the original position
	center->x += oldeye.x;
	center->y += oldeye.y;
	center->z += oldeye.z;
	up->x += oldeye.x;
	up->y += oldeye.y;
	up->z += oldeye.z;
	side->x += oldeye.x;
	side->y += oldeye.y;
	side->z += oldeye.z;
	eye->x += oldeye.x; //
	eye->y += oldeye.y; // :)
	eye->z += oldeye.z; //
}

void TurnUp(float degrees, Vector3D *eye, Vector3D *center, Vector3D *up, Vector3D *side)
{
	//In Crystal Ball interface:
	//vec3 axis = normalize(cross(eye, up));
	//eye = eye * rotate(degrees, axis);
	//up = up * rotate(degrees, axis);
	/*
	Vector3D axis = Normalize(CrossProduct(*eye, *up));
	*eye = TransformVector(*eye, Rotate(degrees, axis));
	*up = TransformVector( *up, Rotate(degrees, axis));
	*/

	//Set up axis
	Vector3D sideaxis;
	sideaxis.x = side->x - eye->x;
	sideaxis.y = side->y - eye->y;
	sideaxis.z = side->z - eye->z;
	sideaxis = Normalize(sideaxis);

	Vector3D oldeye;
	oldeye.x = eye->x;
	oldeye.y = eye->y;
	oldeye.z = eye->z;

	//Translate to world origin
	center->x -= eye->x;
	center->y -= eye->y;
	center->z -= eye->z;
	up->x -= eye->x;
	up->y -= eye->y;
	up->z -= eye->z;
	side->x -= eye->x;
	side->y -= eye->y;
	side->z -= eye->z;
	eye->x -= eye->x; //
	eye->y -= eye->y; // :)
	eye->z -= eye->z; //

	//Transform (rotate)
	*eye = TransformVector(*eye, Rotate(degrees, sideaxis));
	*center = TransformVector(*center, Rotate(degrees, sideaxis));
	*up = TransformVector(*up, Rotate(degrees, sideaxis));
	*side = TransformVector(  *side, Rotate(degrees, sideaxis));

	//Translate back
	center->x += oldeye.x;
	center->y += oldeye.y;
	center->z += oldeye.z;
	up->x += oldeye.x;
	up->y += oldeye.y;
	up->z += oldeye.z;
	side->x += oldeye.x;
	side->y += oldeye.y;
	side->z += oldeye.z;
	eye->x += oldeye.x;
	eye->y += oldeye.y;
	eye->z += oldeye.z;
}

void RollLeft(float degrees, Vector3D *eye, Vector3D *center, Vector3D *up, Vector3D *side)
{
	//In Crystal Ball interface:
	//vec3 axis = normalize(center - eye);
	//up = up * rotate(degrees, axis);
	/*
	Vector3D axis;
	axis.x = center.x - eye.x;
	axis.y = center.y - eye.y;
	axis.z = center.z - eye.z;
	axis = Normalize(axis);
	*up = TransformVector(*up, Rotate(degrees, axis));
	*/

	//Set up rotation axis
	Vector3D frontaxis;
	frontaxis.x = center->x - eye->x;
	frontaxis.y = center->y - eye->y;
	frontaxis.z = center->z - eye->z;
	frontaxis = Normalize(frontaxis);

	Vector3D oldeye;
	oldeye.x = eye->x;
	oldeye.y = eye->y;
	oldeye.z = eye->z;

	//Translate to world origin
	center->x -= eye->x;
	center->y -= eye->y;
	center->z -= eye->z;
	up->x -= eye->x;
	up->y -= eye->y;
	up->z -= eye->z;
	side->x -= eye->x;
	side->y -= eye->y;
	side->z -= eye->z;
	eye->x -= eye->x;
	eye->y -= eye->y;
	eye->z -= eye->z;

	//Transform (rotate)
	*eye = TransformVector(*eye, Rotate(degrees, frontaxis));
	*center = TransformVector(*center, Rotate(degrees, frontaxis));
	*up = TransformVector(*up, Rotate(degrees, frontaxis));
	*side = TransformVector(  *side, Rotate(degrees, frontaxis));

	//Translate back
	center->x += oldeye.x;
	center->y += oldeye.y;
	center->z += oldeye.z;
	up->x += oldeye.x;
	up->y += oldeye.y;
	up->z += oldeye.z;
	side->x += oldeye.x;
	side->y += oldeye.y;
	side->z += oldeye.z;
	eye->x += oldeye.x;
	eye->y += oldeye.y;
	eye->z += oldeye.z;
}

typedef void (*MovementCallback)( GLfloat amount
			, Vector3D* eye
			, Vector3D* center
			, Vector3D* up
			, Vector3D* side);

void Move(HWND Handle, HGLRC hrc, GLfloat move_amount, Camera* camera, MovementCallback Movement)
{
	const int number_of_move_fractions = 10;
	GLfloat move_amount_inc = move_amount / number_of_move_fractions;
	GLfloat abs_move_amount = fabs(move_amount);
	GLfloat abs_move_amount_inc = fabs(move_amount_inc);
	for (GLfloat move_amount_tmp = 0.0f; move_amount_tmp <= abs_move_amount; move_amount_tmp += abs_move_amount_inc)
	{
		Movement(move_amount_inc, &(camera->eye), &(camera->center), &(camera->up), &(camera->side));
		FormPaint(Handle, hrc);
	}
}

void VoxelLaserKill();


void FormKeyDown(HWND Handle, HGLRC hglrc, WPARAM Key)
{
	if(Key == VK_ESCAPE)
		exit(0);
	if (Key == VK_LEFT)
	{
		Move(Handle, hglrc, degleft, player, TurnLeft);
	}
	if (Key == VK_RIGHT)
	{
		Move(Handle, hglrc, -degleft, player, TurnLeft);
	}
	if (Key == VK_UP)
	{
		Move(Handle, hglrc, degup, player, TurnUp);
	}
	if (Key == VK_DOWN)
	{
		Move(Handle, hglrc, -degup, player, TurnUp);
	}
	if (Key == VK_SPACE)
	{
		Move(Handle, hglrc, stepy, player, MoveY);
	}
	if (Key == VK_CONTROL)
	{
		Move(Handle, hglrc, -stepy, player, MoveY);
	}
	if (Key == 'W')
	{
		Move(Handle, hglrc, stepz, player, MoveZ);
	}
	if (Key == 'S')
	{
		Move(Handle, hglrc, -stepz, player, MoveZ);
	}
	if (Key == 'A')
	{
		Move(Handle, hglrc, -stepx, player, MoveX);
	}
	if (Key == 'D')
	{
		Move(Handle, hglrc, stepx, player, MoveX);
	}
	if (Key == 'O')
	{
		Move(Handle, hglrc, -degroll, player, RollLeft);
	}
	if (Key == 'P')
	{
		Move(Handle, hglrc, degroll, player, RollLeft);
	}
	if (Key == 'U')
	{
		// TODO: MOUSE LOOK
	}
	if (Key == 'X')
	{
		fire = true;
		//FormPaint();
	}
	if (Key == 'X')
	{
		fire = false;
		VoxelLaserKill();
	}

	InvalidateRect(Handle, NULL, false);
}

void destroy(HWND Handle, HDC DC, HGLRC hrc )
{
	DeleteScreen(&screen);
	wglMakeCurrent(NULL, NULL);
	wglDeleteContext(hrc);
	ReleaseDC(Handle, DC);
	DeleteDC(DC);
}


struct Voxel;
typedef void (*ProcessVoxelCallback)(long sx, long sy, float fade, Voxel* voxel, long index);
typedef bool (*VisitVoxelCallback)( long sx, long sy, long x, long y, long z, float fade, ProcessVoxelCallback ProcessVoxel);

Vector3D viewport[3];
Vector3D viewport_dxvector;
Vector3D viewport_dyvector;
Vector3D up;
Vector3D front;
Vector3D right;

struct Point3D
{
	double x, y, z;
	Point3D() : x(0.0), y(0.0), z(0.0) {}
};
Point3D ipos, istr, ihei, ifor;

const long kBitArraySize = 33554432; //1024 * 1024 * 256 / 8
unsigned char bitarray[kBitArraySize];
struct Voxel
{
	unsigned long x, y, z, argb;
	Voxel() : x(0), y(0), z(0), argb(0) {}
};
std::vector<Voxel*> colorarray;

//(Ken Silverman):
//Called at least once for every voxel of the board
//issolid: 0 - air, 1 - solid
void setgeom (long x, long y, long z, long issolid)
{
	long index = (256 * (1024 * x + y) + z);
	if (issolid)
	{
		bitarray[index / 8] |= (1 << (index % 8));
	}
	else
	{
		bitarray[index / 8] &= ~(1 << (index % 8));
	}
}

long getgeom(long x, long y, long z)
{
	long index = (256 * (1024 * x + y) + z);
	return bitarray[index / 8] & (1 << (index % 8));
}

//(Ken Silverman):
//Called only for surface voxels
//A surface voxel is any solid voxel with at least 1 air voxel
//  on one of its 6 sides. All solid voxels at z = 0 are automatically
//  surface voxels, but this is not true for x = 0, x = 1023, y = 0, y = 1023,
//  z = 255 (I believe)
//argb: 32-bit color, high byte is used for shading scale (can be ignored)
bool setcol (long x, long y, long z, long argb)
{
	Voxel* voxel = new Voxel();
	if (NULL == voxel)
	{
		return false;
	}
	voxel->x = x;
	voxel->y = y;
	voxel->z = z;
	voxel->argb = argb;
	long index = 256 * (1024 * x + y) + z;
	colorarray[index] = voxel;
	return true;
}

bool LoadVxl(char* filename)
{
	FILE* file;
	long i = 0;
	unsigned char *v = NULL, *vbuf = NULL;

	memset(bitarray, 0, kBitArraySize);
	if (0 != colorarray.size())
	{
		colorarray.clear();
	}

	file = fopen(filename, "rb");
	if (NULL == file)
	{
		//ShowMessage("LoadVxl() file open fail!\n");
		return false;
	}

	fread(&i, 4, 1, file);
	if (i != 0x09072000)
	{
		//ShowMessage("LoadVxl() signature reading fail!\n");
		fclose(file);
		return false;
	}

	fread(&i, 4, 1, file);
	if (1024 != i)
	{
		//ShowMessage("LoadVxl() X-dimension read fail! Expected 1024, read ");
		fclose(file);
		return false;
	}

	fread(&i, 4, 1, file);
	if (1024 != i)
	{
		//ShowMessage("LoadVxl() Y-dimension read fail! Expected 1024, read ");
		fclose(file);
		return false;
	}

	fread(&ipos, 24, 1, file); //camera position
	fread(&istr, 24, 1, file); //unit right vector
	fread(&ihei, 24, 1, file); //unit down vector
	fread(&ifor, 24, 1, file); //unit forward vector

	i = filelength(_fileno(file)) - ftell(file);
	vbuf = new unsigned char[i];
	if (NULL == vbuf)
	{
		//ShowMessage("LoadVxl() V-buffer allocation fail!\n");
		fclose(file);
		return false;
	}

	fread(vbuf, i, 1, file);
	fclose(file);

	//  1024 * 1024 * 256 * 16 bytes = 4GB
	colorarray.resize(1024 * 1024 * 256);
	for (int i = 0;  i < colorarray.size(); ++i)
	{
		colorarray[i] = NULL;
	}

	memset(bitarray, 0xFF, 33554432);

	v = vbuf;
	for(long y = 0; y < 1024; ++y)
	{
		for(long x = 0; x < 1024; ++x)
		{
			long z = 0;
			while (1)
			{
				for(long i = z; i < v[1]; ++i)
				{
					setgeom(x, y, i, 0);
				}

				for(long z = v[1]; z <= v[2]; ++z)
				{
					bool result = setcol(x, y, z, *(long *)&v[(z - v[1] + 1) << 2]);
					if (false == result)
						return result;
				}

				if (!v[0])
				{
					break;
				}

				z = v[2] - v[1] - v[0] + 2;
				v += v[0] << 2;

				for(z += v[3]; z<v[3]; ++z)
				{
					bool result = setcol(x, y, z, *(long *)&v[(z - v[3]) << 2]);
					if (false == result)
						return result;
				}
			}
			v += ((((long)v[2]) - ((long)v[1]) + 2) << 2);
		}
	}
	delete [] vbuf;
	return true;
}

void PixelSet(const long sx, const long sy, unsigned long argb, float fade, _Screen* screen)
{
	if ((sx < 0) || (sx > screen->width - 1) || (sy < 0) || (sy > screen->height - 1))
	{
		return;
	}

	long index = screen->width * sy + sx;
	unsigned char red =  (argb & 0x000000FF);
	unsigned char grn = ((argb & 0x0000FF00) >> 8);
	unsigned char blu = ((argb & 0x00FF0000) >> 16);

	screen->buffer[4 * index + 0] = red * fade;
	screen->buffer[4 * index + 1] = grn * fade;
	screen->buffer[4 * index + 2] = blu * fade;
	screen->buffer[4 * index + 3] = 255;
}

void DrawVoxelSample(long sx, long sy, float fade, Voxel* voxel, long index)
{
	if (NULL == voxel)
		return;
	PixelSet(sx, sy, voxel->argb, fade, &screen);
}

void ShootVoxel(long sx, long sy, float fade, Voxel* voxel, long index)
{
	if (NULL == voxel)
		return;
	unsigned long voxel_paintcolor = voxel->argb;
	setgeom(voxel->x, voxel->y, voxel->z, 0);

	const long x0 = voxel->x;
	const long y0 = voxel->y;
	const long z0 = voxel->z;
	const long radius = 10;
	for (long x = x0 - 2 * radius; x < x0 + 2 * radius; ++x)
	{
		for (long y = y0 - 2 * radius; y < y0 + 2 * radius; ++y)
		{
			for (long z = voxel->z - 2 * radius; z < voxel->z + 2 * radius; ++z)
			{
				if ((x < 0) || (x >= 1024) || (y < 0) || (y >= 1024) ||
					(z < 0) && (z >= 256))
				{
					continue;
				}

				long current_index = 256 * (1024 * x + y) + z;

				if (current_index == index)
					continue;

				long sphere = (x - x0) * (x - x0) +
					(y - y0) * (y - y0) +
					(z - z0) * (z - z0);

				if (sphere < radius * radius)
				{
					setgeom(x, y, z, 0);
					if (NULL != colorarray[current_index])
					{
						delete colorarray[current_index];
						colorarray[current_index] = NULL;
					}
				}
				else if ((sphere >= radius * radius) &&
							(sphere <= (radius + radius / 4) *
									(radius + radius / 4)))
				{
					// TODO: Fill gaps
					//setgeom(x, y, z, 1);
					//setcol(x, y, z, voxel_paintcolor);

					// Shell variant
					if (0 != getgeom(x, y, z))
					{
						Voxel* current_voxel = colorarray[current_index];
						if (NULL != current_voxel)
						{
							current_voxel->argb = voxel_paintcolor;
						}
					}
				}
			}
		}
	}

	colorarray[index] = NULL;
	if (NULL != voxel)
	{
		delete voxel;
		voxel = NULL;
	}
}

bool VisitVoxel( long sx, long sy, long x, long y, long z, float fade, ProcessVoxelCallback ProcessVoxel)
{
	if ((x < 0) || (x >= 1024) || (y < 0) || (y >= 1024) || (z < 0) && (z >= 256))
	{
		return false;
	}

	if (0 == getgeom(x, y, z))
		return false;
	long index = 256 * (1024 * x + y) + z;

	if (NULL == colorarray[index])
		return false;
	Voxel* voxel = colorarray[index];
	ProcessVoxel(sx, sy, fade, voxel, index);
	return true;
}

long sgn(long x)
{
	if (0 == x)
		return 1;
	return x / abs(x);
}

void Line( long scrx, long scry, long x, long y, long z
	, long dx, long dy, long dz
	, VisitVoxelCallback VisitVoxel
	, ProcessVoxelCallback ProcessVoxel)
{
	long n, sx, sy, sz, exy, exz, ezy, ax, ay, az, bx, by, bz;

	sx = sgn(dx); sy = sgn(dy); sz = sgn(dz);
	ax = abs(dx); ay = abs(dy); az = abs(dz);
	bx = 2 * ax;  by = 2 * ay;  bz = 2 * az;
	exy = ay - ax;
	exz = az - ax;
	ezy = ay - az;

	bool hit = false;
	long dist = ax + ay + az;
	n = dist;
	while (--n)
	{
		float fade = (float)n / (float)dist;
		bool hit = VisitVoxel(scrx, scry, x, y, z, fade, ProcessVoxel);
		if (true == hit) break;
		if (exy < 0)
		{
			if (exz < 0)
			{
				x += sx;
				exy += by;
				exz += bz;
			}
			else
			{
				z += sz;
				exz -= bx;
				ezy += by;
			}
		}
		else
		{
			if (ezy < 0)
			{
			z += sz;
			exz -= bx;
			ezy += by;
			}
			else
			{
			y += sy;
			exy -= bx;
			ezy -= bz;
			}
		}
	}
}

void SetupLCS(const Camera& camera, Vector3D* up, Vector3D* front, Vector3D* right)
{
	up->x = camera.up.x - camera.eye.x;
	up->y = camera.up.y - camera.eye.y;
	up->z = camera.up.z - camera.eye.z;

	front->x = camera.center.x - camera.eye.x;
	front->y = camera.center.y - camera.eye.y;
	front->z = camera.center.z - camera.eye.z;

	right->x = camera.side.x - camera.eye.x;
	right->y = camera.side.y - camera.eye.y;
	right->z = camera.side.z - camera.eye.z;
}

void SetupViewport( const Camera& camera
	, const Vector3D& up, const Vector3D& front, const Vector3D& right
	, const GLfloat d1, const GLfloat d2
	, Vector3D viewport[3]
	, Vector3D* viewport_dxvector, Vector3D* viewport_dyvector)
{
	viewport[0].x = camera.eye.x + d1 * (up.x - right.x);
	viewport[0].y = camera.eye.y + d1 * (up.y - right.y);
	viewport[0].z = camera.eye.z + d1 * (up.z - right.z);

	viewport[1].x = camera.eye.x + d1 * (up.x + right.x);
	viewport[1].y = camera.eye.y + d1 * (up.y + right.y);
	viewport[1].z = camera.eye.z + d1 * (up.z + right.z);

	viewport[2].x = camera.eye.x - d1 * (up.x + right.x);
	viewport[2].y = camera.eye.y - d1 * (up.y + right.y);
	viewport[2].z = camera.eye.z - d1 * (up.z + right.z);

	for (int i = 0; i < 3; ++i)
	{
		viewport[i].x += d2 * front.x;
		viewport[i].y += d2 * front.y;
		viewport[i].z += d2 * front.z;
		viewport[i].w = 1.0f;
	}

	viewport_dxvector->x = viewport[1].x - viewport[0].x;
	viewport_dxvector->y = viewport[1].y - viewport[0].y;
	viewport_dxvector->z = viewport[1].z - viewport[0].z;
	viewport_dxvector->w = 1.0f;

	viewport_dyvector->x = viewport[2].x - viewport[0].x;
	viewport_dyvector->y = viewport[2].y - viewport[0].y;
	viewport_dyvector->z = viewport[2].z - viewport[0].z;
	viewport_dyvector->w = 1.0f;
}

bool CreateScreen(const int width, const int height, _Screen* screen)
{
	screen->width = width;
	screen->height = height;
	screen->buffer = new unsigned char[4 * width * height];
	if (NULL == screen->buffer)
		return false;
	memset(screen->buffer, 0, 4 * width * height);
	glEnable(GL_TEXTURE_2D);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
	glGenTextures(1, &(screen->texture));
	glBindTexture(GL_TEXTURE_2D, screen->texture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, screen->buffer);
	return true;
}

void DeleteScreen(_Screen* screen)
{
	if (NULL != screen->buffer)
	{
		delete [] screen->buffer;
		screen->buffer = NULL;
	}
	glDeleteTextures(1, &(screen->texture));
}

bool FlipScreen(const _Screen screen)
{
	if (NULL == screen.buffer)
		return false;
	glBindTexture(GL_TEXTURE_2D, screen.texture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, screen.width, screen.height, 0, GL_RGBA, GL_UNSIGNED_BYTE, screen.buffer);
	return true;
}

void DisplayScreen(const _Screen screen)
{
	glBindTexture(GL_TEXTURE_2D, screen.texture);
	glColor3f(1.0, 1.0, 1.0);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);
		glVertex3f(-1.0,  1.0, 0.0);
		glTexCoord2f(1.0, 0.0);
		glVertex3f( 1.0,  1.0, 0.0);
		glTexCoord2f(1.0, 1.0);
		glVertex3f( 1.0, -1.0, 0.0);
		glTexCoord2f(0.0, 1.0);
		glVertex3f(-1.0, -1.0, 0.0);
	glEnd();
}

bool ClearScreen(_Screen* screen)
{
	if (NULL == screen->buffer)
		return false;
	memset(screen->buffer, 0, 4 * screen->width * screen->height);
}

void DrawViewport()
{
	glLineWidth(3.0);
	glColor3f(0.0, 0.0, 0.0);
	glBegin(GL_LINES);
		glVertex3f(viewport[0].x, viewport[0].y, viewport[0].z);
		glVertex3f(viewport[1].x, viewport[1].y, viewport[1].z);
		glVertex3f(viewport[0].x, viewport[0].y, viewport[0].z);
		glVertex3f(viewport[2].x, viewport[2].y, viewport[2].z);
	glEnd();
}

void VoxelLaserKill()
{
	SetupLCS(*player, &up, &front, &right);

	Vector3D raystart, rayend;
	raystart.x = player->eye.x;
	raystart.y = player->eye.y;
	raystart.z = player->eye.z;
	raystart.w = 1.0f;
	rayend.x = raystart.x + front.x;
	rayend.y = raystart.y + front.y;
	rayend.z = raystart.z + front.z;
	rayend.w = 1.0f;

	Voxel voxel_raystart, voxel_rayend, voxel_ray;
	voxel_raystart.x = (raystart.x - boardorigin.x) / h;
	voxel_raystart.y = (raystart.y - boardorigin.y) / h;
	voxel_raystart.z = (raystart.z - boardorigin.z) / h;
	voxel_rayend.x = (rayend.x - boardorigin.x) / h;
	voxel_rayend.y = (rayend.y - boardorigin.y) / h;
	voxel_rayend.z = (rayend.z - boardorigin.z) / h;
	voxel_ray.x = voxel_rayend.x - voxel_raystart.x;
	voxel_ray.y = voxel_rayend.y - voxel_raystart.y;
	voxel_ray.z = voxel_rayend.z - voxel_raystart.z;
	voxel_ray.x *= kRange;
	voxel_ray.y *= kRange;
	voxel_ray.z *= kRange;

	Line( screen_width / 2, screen_height / 2
		, voxel_raystart.x, voxel_raystart.y, voxel_raystart.z
		, voxel_ray.x, voxel_ray.y, voxel_ray.z
		, VisitVoxel, ShootVoxel);
}

void DrawVxlRaycast()
{
	if (NULL == screen.buffer)
		return;

	SetupLCS(*player, &up, &front, &right);
	SetupViewport(*player, up, front, right, d1, d2, viewport, &viewport_dxvector, &viewport_dyvector);

	Vector3D raystart;
	raystart.x = player->eye.x;
	raystart.y = player->eye.y;
	raystart.z = player->eye.z;
	raystart.w = 1.0f;

	Voxel voxel_raystart;
	voxel_raystart.x = (raystart.x - boardorigin.x) / h;
	voxel_raystart.y = (raystart.y - boardorigin.y) / h;
	voxel_raystart.z = (raystart.z - boardorigin.z) / h;


	memset(screen.buffer, 0, 4 * screen.width * screen.height);
	for (long x = 0; x < screen_width; ++x)
	{
		for (long y = 0; y < screen_height; ++y)
		{
			Vector3D rayend;
			rayend.x = viewport[0].x + x * ddx * viewport_dxvector.x +
									+ y * ddy * viewport_dyvector.x;
			rayend.y = viewport[0].y + x * ddx * viewport_dxvector.y +
									+ y * ddy * viewport_dyvector.y;
			rayend.z = viewport[0].z + x * ddx * viewport_dxvector.z +
									+ y * ddy * viewport_dyvector.z;
			rayend.w = 1.0f;

			Voxel voxel_rayend;
			voxel_rayend.x = (rayend.x - boardorigin.x) / h;
			voxel_rayend.y = (rayend.y - boardorigin.y) / h;
			voxel_rayend.z = (rayend.z - boardorigin.z) / h;

			Voxel voxel_ray;
			voxel_ray.x = voxel_rayend.x - voxel_raystart.x;
			voxel_ray.y = voxel_rayend.y - voxel_raystart.y;
			voxel_ray.z = voxel_rayend.z - voxel_raystart.z;
			voxel_ray.x *= kFar;
			voxel_ray.y *= kFar;
			voxel_ray.z *= kFar;
			Line(x, y, voxel_raystart.x, voxel_raystart.y, voxel_raystart.z
					, voxel_ray.x, voxel_ray.y, voxel_ray.z
					, VisitVoxel, DrawVoxelSample);
		}
	}
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, screen.texture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, screen.width, screen.height, 0, GL_RGBA, GL_UNSIGNED_BYTE, screen.buffer);
	glBindTexture(GL_TEXTURE_2D, screen.texture);
	glColor3f(1.0, 1.0, 1.0);

	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);
		glVertex3f(-1.0,  1.0, 0.0);
		glTexCoord2f(1.0, 0.0);
		glVertex3f( 1.0,  1.0, 0.0);
		glTexCoord2f(1.0, 1.0);
		glVertex3f( 1.0, -1.0, 0.0);
		glTexCoord2f(0.0, 1.0);
		glVertex3f(-1.0, -1.0, 0.0);
	glEnd();
}
